// Explosion test program (c)1998 James Shaughnessy
//
// -----------------------------------------
// James Shaughnessy    james@manc.u-net.com
// http://www.netyaroze-europe.com/~shaughnj
// -----------------------------------------

#include <libps.h>
#include <rand.h>
#include "pad.h"

// DEBUG flag
//#define DEBUG


#define FIREBALL_ADDR (0x80090000)

#define OT_LENGTH (1)
#define NUM_SPRITES (512)
#define NUM_PARTICLES (NUM_SPRITES/2)
#define EXPLODE_TIME (256)
#define GRAVITY_FORCE (256)
#define AIR_RESISTANCE (64)


typedef struct {
  long x, y;                            // coords *4096
  short vx, vy;                         // velocity
  short rotation;
  short floor;                          // "floor" of explosion
  GsSPRITE sprite;                      // it's actual sprite
} GsPARTICLE;

typedef struct {
  short x, y;                           // coords
  u_long timer;
  GsPARTICLE particle[NUM_PARTICLES];   // list of it's particles
} GsEXPLOSION;



// Global Variables
int activeBuff;
int hsyncs;
u_long timer = 0;
u_long padStatus = 0;
extern DISPENV GsDISPENV;

// Two Ordering Tables, one for each buffer
GsOT WorldOT[2];
// Two Ordering Table Tags, one for each buffer
GsOT_TAG OTTag[2][1<<OT_LENGTH];
//Also allocate memory used for depth sorting, a block for each buffer
PACKET GpuPacketArea[2][NUM_SPRITES*sizeof(GsSPRITE)];

GsEXPLOSION Explosion;


// Function Prototypes
void InitGraphics();
void LoadTIM(u_long memaddr);
void InitExplosion(short floor);
void Explode();



void main()
{
  PadInit();
  InitGraphics();

  while(1)
  {
    padStatus = PadRead();

    // Quit if both Select and Start pressed
    if (padStatus & PADselect && padStatus & PADstart) break;

    activeBuff = GsGetActiveBuff();

    // Set packet work base
    GsSetWorkBase((PACKET *)GpuPacketArea[activeBuff]);

    // Clear the Ordering Table
    GsClearOt(0, 0, &WorldOT[activeBuff]);


    if (Explosion.timer)
      Explode();
    else if (padStatus & PADcross)
      InitExplosion(256);

#ifdef DEBUG
    // Display H-Sync counter
    FntPrint("HSyncs: %d\n", hsyncs);
    FntFlush(-1);
#endif

    // Wait for end of drawing, a vertical blank then swap buffers
    DrawSync(0);
    hsyncs = VSync(0);
    GsSwapDispBuff();

    // Register Clear in OT
    GsSortClear(0, 0, 0, &WorldOT[activeBuff]);

    // Draw the Ordering Table
    GsDrawOt(&WorldOT[activeBuff]);

    timer++;
  }

  ResetGraph(0);
}


void Explode()
{
  int i;
#ifdef DEBUG
  FntPrint("EXPLODING.  Like, BOOM or something.\n");
#endif

  Explosion.timer--;


  for (i=0; i<NUM_PARTICLES; i++) {
    // Update particle location
    Explosion.particle[i].vx -= (Explosion.particle[i].vx*AIR_RESISTANCE)>>12;
    Explosion.particle[i].vy -= ((Explosion.particle[i].vy*AIR_RESISTANCE)>>12) - GRAVITY_FORCE;
    Explosion.particle[i].x += Explosion.particle[i].vx;
    Explosion.particle[i].y += Explosion.particle[i].vy;
    Explosion.particle[i].sprite.x = (Explosion.particle[i].x>>12);
    Explosion.particle[i].sprite.y = (Explosion.particle[i].y>>12);
    Explosion.particle[i].sprite.rotate += Explosion.particle[i].rotation;
    // Bounce off "floor"
    if (Explosion.particle[i].sprite.y > Explosion.particle[i].floor)
    {
      Explosion.particle[i].y -= Explosion.particle[i].vy;
      Explosion.particle[i].vy = -Explosion.particle[i].vy;
    }
    // Sort particle sprites in OT
    GsSortSprite(&Explosion.particle[i].sprite, &WorldOT[activeBuff], 0);

  }
}

// Inits explosion data
void InitExplosion(short floor)
{
  int i;

  LoadTIM(FIREBALL_ADDR);
  srand(timer);

  Explosion.x = 160;
  Explosion.y = 128;
  Explosion.timer = EXPLODE_TIME;
  for (i=0; i<NUM_PARTICLES; i++) {
    Explosion.particle[i].x = (Explosion.x<<12);
    Explosion.particle[i].y = (Explosion.y<<12);
    Explosion.particle[i].vx = ((rand()%256) - 128)<<8;
    Explosion.particle[i].vy = ((rand()%256) - 200)<<8;
    Explosion.particle[i].rotation = ((rand()%256) - 128)<<14;
    Explosion.particle[i].floor = floor;
    Explosion.particle[i].sprite.attribute = 0;
    Explosion.particle[i].sprite.x = Explosion.x;
    Explosion.particle[i].sprite.y = Explosion.y;
    Explosion.particle[i].sprite.w = 12;
    Explosion.particle[i].sprite.h = 9;
    Explosion.particle[i].sprite.tpage = GetTPage(0, 0, 320, 0);
    Explosion.particle[i].sprite.u = 12 * (rand()%3);
    Explosion.particle[i].sprite.v = 0;
    Explosion.particle[i].sprite.cx = 320;
    Explosion.particle[i].sprite.cy = 9;
    Explosion.particle[i].sprite.r =
    Explosion.particle[i].sprite.g =
    Explosion.particle[i].sprite.b = 128;
    Explosion.particle[i].sprite.mx = 5;
    Explosion.particle[i].sprite.my = 4;
    Explosion.particle[i].sprite.scalex =
    Explosion.particle[i].sprite.scaley = ONE;
    Explosion.particle[i].sprite.rotate = 0;
  }
}


// Initialises Video and the Frame Buffer
void InitGraphics()
{
  int i;

  // Setup graphics PAL 320x256 
  SetVideoMode(MODE_PAL);
  GsInitGraph(320, 256, 4, 0, 0);
  GsDISPENV.screen.y = 20;
  GsDISPENV.screen.h = 256;
  GsDefDispBuff(0, 0, 0, 256);

  // Init Ordering Tables
  for (i = 0; i < 2; i++) {
    WorldOT[i].length = OT_LENGTH;
    WorldOT[i].org = OTTag[i];
    GsClearOt(0, 0, &WorldOT[i]);
  }

  // Loads the font into Frame Buffer
  FntLoad(960, 256);
  // Font printing location
  FntOpen(16, 16, 304, 240, 0, 512);
}


// Loads a TIM into the frame buffer
void LoadTIM(u_long memaddr)
{
  RECT rect;
  GsIMAGE im;

  // Get TIM info
  GsGetTimInfo((u_long *)memaddr+1, &im);

  // Set rect and load into frame buffer
  setRECT(&rect, im.px, im.py, im.pw, im.ph);
  LoadImage(&rect, im.pixel);

  // Test if TIM has a clut (4 or 8 bit) and so load clut too
  if ((im.pmode>>3) & 1)
  {
    setRECT(&rect, im.cx, im.cy, im.cw, im.ch);
    LoadImage(&rect, im.clut);
  }

  DrawSync(0);
}


// ****** PAD routines ******

// call once only in program initialisation
void PadInit (void)
{
  GetPadBuf(&bb0, &bb1);
}

// call once per VSync(0)
// puts controller pad status into unsigned long integer
// please refer to the manuals if you want explanation 
// of the internals of this function
u_long PadRead(void)
{
  return(~(*(bb0+3) | *(bb0+2) << 8 | *(bb1+3) << 16 | *(bb1+2) << 24));
}

